/* 
 * Copyright (C) 2004      Disch
 * Copyright (C) 2006      hcs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  NSF_Core.h
//

union TWIN
{
	WORD						W;
	struct{ BYTE l; BYTE h; }	B;
};

union QUAD
{
	UINT								D;
	struct{ BYTE l; BYTE h; WORD w; }	B;
};

struct NSF_ADVANCEDOPTIONS
{
	int			nSilenceTrackMS;
	int			nInvertCutoffHz;
	BYTE		bDMCPopReducer;
	BYTE		nForce4017Write;
	BYTE		bN106PopReducer;
	BYTE		bFDSPopReducer;

	BYTE		bIgnore4011Writes;
	BYTE		bIgnoreBRK;
	BYTE		bIgnoreIllegalOps;
	BYTE		bNoWaitForReturn;

	BYTE		bPALPreference;
	BYTE		bCleanAXY;
	BYTE		bResetDuty;
	BYTE		bNoSilenceIfTime;

	BYTE		bHighPassEnabled;
	BYTE		bLowPassEnabled;
	BYTE		bPrePassEnabled;
	int			nHighPassBase;
	int			nLowPassBase;
	int			nPrePassBase;
};

#define NTSC_FREQUENCY			 1789772.727273f
#define PAL_FREQUENCY			 1652097.692308f
#define NTSC_NMIRATE			      60.098814f
#define PAL_NMIRATE				      50.006982f

#define NES_FREQUENCY				21477270
#define NTSC_FRAME_COUNTER_FREQ		(NTSC_FREQUENCY / (NES_FREQUENCY / 89490.0f))
#define PAL_FRAME_COUNTER_FREQ		(PAL_FREQUENCY / (NES_FREQUENCY / 89490.0f))


#define EXTSOUND_VRC6			0x01
#define EXTSOUND_VRC7			0x02
#define EXTSOUND_FDS			0x04
#define EXTSOUND_MMC5			0x08
#define EXTSOUND_N106			0x10
#define EXTSOUND_FME07			0x20

#include <unistd.h>
#include <string.h>
#include <math.h>

struct NSFFile;

typedef BYTE ( *ReadProc)(WORD);
typedef void ( *WriteProc)(WORD,BYTE);


int		NSFCore_Initialize();				//1 = initialized ok, 0 = couldn't initialize (memory allocation error)
void	NSFCore_Destroy();					//cleans up memory

//
//	Song Loading
//
int		LoadNSF(void);	//grab data from an existing file  1 = loaded ok, 0 = error loading

//
//	Track Control
//
void	SetTrack(BYTE track);			//Change tracks

//
//	Getting Samples
//
int		GetSamples(BYTE* buffer, int buffersize);	//fill a buffer with samples

//
//	Playback options
//
int		SetPlaybackOptions(int samplerate,int channels);				//Set desired playback options (0 = bad options couldn't be set)
void	SetPlaybackSpeed(float playspersec);							//Speed throttling (0 = uses NSF specified speed)
void	SetMasterVolume(float vol);										//1 = full volume, 2 = double volume, .5 = half volume, etc
void	SetChannelOptions(UINT chan,int mix,int vol,int pan,int inv);	//chan is the zero based index of the channel to change.
void	SetAdvancedOptions(const struct NSF_ADVANCEDOPTIONS* opt);				//misc options

float	GetPlaybackSpeed();
float	GetMasterVolume();
void	GetAdvancedOptions(struct NSF_ADVANCEDOPTIONS* opt);

//
//	Seeking
//
float	GetPlayCalls();								//gets the number of 'play' routine calls executed
UINT	GetWrittenTime(float basedplayspersec);	//gets the output time (based on the given play rate, if basedplayspersec is zero, current playback speed is used
void	SetPlayCalls(float plays);					//sets the number os 'plays' routines executed (for precise seeking)
void	SetWrittenTime(UINT ms,float basedplays);	//sets the written time (approx. seeking)

//
//	Fading
//

void	StopFade();									//stops all fading (plays indefinitely)
BYTE	SongCompleted();							//song has faded out (samples have stopped being generated)
void	SetFade(int fadestart,int fadestop,BYTE bNotDefault);						//parameters are play calls
void	SetFadeTime(UINT fadestart,UINT fadestop,float basedplays,BYTE bNotDefault);//parameters are in milliseconds

//protected:
//
//	Internal Functions
//
void	RebuildOutputTables(UINT chans);			//Rebuilds the output tables (when output options are changed.. like volume, pan, etc)
void	CalculateChannelVolume(int maxvol,int * left, int * right, BYTE vol, char pan);	//used in above function
void	RecalculateFade();							//called when fade status is changed.
void	RecalcFilter();
void	RecalcSilenceTracker();
void	RecalculateInvertFreqs(int cutoff);


void 		WriteMemory_VRC6(WORD a,BYTE v);
void 		WriteMemory_MMC5(WORD a,BYTE v);
void 		WriteMemory_N106(WORD a,BYTE v);
void 		WriteMemory_VRC7(WORD a,BYTE v);
void 		WriteMemory_FME07(WORD a,BYTE v);

void		VRC7_Init();
void		VRC7_Reset();
void		VRC7_Mix();
void		VRC7_Write(BYTE val);
void		VRC7_Destroy();
void		VRC7_LoadInstrument(BYTE Chan);
void		VRC7_RecalcMultiplier(BYTE chan);
void		VRC7_ChangeInversion(BYTE chan,BYTE inv);
void		VRC7_ChangeInversionFreq();

/*
 *	Emulation
 */
void				EmulateAPU(BYTE bBurnCPUCycles);
UINT				Emulate6502(UINT runto);


extern BYTE		regA;
extern BYTE		regX;
extern BYTE		regY;
extern BYTE		regP;
extern BYTE		regSP;
extern WORD		regPC;

extern BYTE		bPALMode;
extern BYTE		bCPUJammed;

extern BYTE		bIgnoreBRK;
extern BYTE		bIgnoreIllegalOps;
extern BYTE		bNoWaitForReturn;
extern BYTE		bPALPreference;
extern BYTE		bCleanAXY;
extern BYTE		bResetDuty;
extern UINT		nCPUCycle;

extern ReadProc	ReadMemory[0x10];
extern WriteProc	WriteMemory[0x10];

extern BYTE 		pRAM[0x800];
extern BYTE*		pStack;

/* stuff needed by VRC7 */
extern BYTE		nVRC7Address;
extern BYTE*		pVRC7Buffer;
extern void*		pFMOPL;
extern BYTE		VRC7Chan[3][6];
extern BYTE		bVRC7_FadeChanged;
extern BYTE		bVRC7Inv[6];

extern int				nSampleRate;
extern int				nMonoStereo;

extern float			fMasterVolume;

extern UINT			nStartFade;
extern UINT			nEndFade;
extern BYTE			bFade;
extern float			fFadeVolume;
extern float			fFadeChange;

extern BYTE		bChannelMix[24];
extern BYTE		nChannelVol[29];
extern char		nChannelPan[29];
extern BYTE*		pOutput;
extern int			nInvertCutoffHz;
