/* 
 * Copyright (C) 2004      Disch
 * Copyright (C) 2006      hcs
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_Square.h
//

struct Wave_Squares
{

	///////////////////////////////////
	// Programmable Timer
	union TWIN	nFreqTimer[2];
	int			nFreqCount[2];

	///////////////////////////////////
	// Length Counter
	BYTE		nLengthCount[2];
	BYTE		bLengthEnabled[2];
	BYTE		bChannelEnabled[2];

	///////////////////////////////////
	// Volume / Decay
	BYTE		nVolume[2];
	BYTE		nDecayVolume[2];
	BYTE		bDecayEnable[2];
	BYTE		bDecayLoop[2];
	BYTE		nDecayTimer[2];
	BYTE		nDecayCount[2];

	///////////////////////////////////
	// Sweep Unit
	BYTE		bSweepEnable[2];
	BYTE		bSweepMode[2];
	BYTE		bSweepForceSilence[2];
	BYTE		nSweepTimer[2];
	BYTE		nSweepCount[2];
	BYTE		nSweepShift[2];

	///////////////////////////////////
	// Duty Cycle
	BYTE		nDutyCount[2];
	BYTE		nDutyCycle[2];

	///////////////////////////////////
	// Output and Downsampling
	BYTE		bChannelMix[2];
	short		nOutputTable_L[0x100];
	short		nOutputTable_R[3][0x100];
	int			nMixL;
	int			nMixR;

	///////////////////////////////////
	// Inverting
	BYTE		bInvert;
	BYTE		bDoInvert;
	WORD		nInvertFreqCutoff;
};

extern struct Wave_Squares	mWave_Squares;			//Square channels 1 and 2

#define this mWave_Squares

inline void Wave_Squares_ClockMajor()		//decay
{
	static int i = 0;
	for(i = 0; i < 2; i++)
	{
		if(this.nDecayCount[i])
			this.nDecayCount[i]--;
		else
		{
			this.nDecayCount[i] = this.nDecayTimer[i];
			if(this.nDecayVolume[i])
				this.nDecayVolume[i]--;
			else
			{
				if(this.bDecayLoop[i])
					this.nDecayVolume[i] = 0x0F;
			}

			if(this.bDecayEnable[i])
				this.nVolume[i] = this.nDecayVolume[i];
		}
	}
}

inline void Wave_Squares_CheckSweepForcedSilence(register int i)
{
	if(this.nFreqTimer[i].W < 8){	this.bSweepForceSilence[i] = 1; return; }
	if(!this.bSweepMode[i] && (( this.nFreqTimer[i].W + (this.nFreqTimer[i].W >> this.nSweepShift[i])) >= 0x0800))
		{ this.bSweepForceSilence[i] = 1; return; }

	this.bSweepForceSilence[i] = 0;
}

inline void Wave_Squares_ClockMinor()		//sweep / length
{
	static int i = 0;
	for(i = 0; i < 2; i++)
	{
		if(this.bLengthEnabled[i] && this.nLengthCount[i])
			this.nLengthCount[i]--;

		if(!this.bSweepEnable[i] || !this.nLengthCount[i] || this.bSweepForceSilence[i] || !this.nSweepShift[i])
			continue;

		if(this.nSweepCount[i])
			this.nSweepCount[i]--;
		else
		{
			this.nSweepCount[i] = this.nSweepTimer[i];
			if(this.bSweepMode[i])	this.nFreqTimer[i].W -= (this.nFreqTimer[i].W >> this.nSweepShift[i]) + !i;
			else				this.nFreqTimer[i].W += (this.nFreqTimer[i].W >> this.nSweepShift[i]);

			Wave_Squares_CheckSweepForcedSilence(i);
		}
	}
}

inline void Wave_Squares_DoTicks(int ticks)
{
	register int mn;
	register BYTE out;

	while(ticks)
	{
		mn = min(this.nFreqCount[0],this.nFreqCount[1]);
		mn = min(mn,ticks);
		ticks -= mn;

		this.nFreqCount[0] -= mn;
		this.nFreqCount[1] -= mn;

		if((this.nDutyCount[0] < this.nDutyCycle[0]) && this.nLengthCount[0] && !this.bSweepForceSilence[0] && this.bChannelMix[0])
			out = (this.nVolume[0] << 4);
		else
			out = 0;

		if((this.nDutyCount[1] < this.nDutyCycle[1]) && this.nLengthCount[1] && !this.bSweepForceSilence[1] && this.bChannelMix[1])
			out |= this.nVolume[1];

		this.nMixL += this.nOutputTable_L[out] * mn;


		if(this.bDoInvert == 3)	this.nMixR -= this.nOutputTable_R[0][out] * mn;
		else				this.nMixR += this.nOutputTable_R[this.bDoInvert][out] * mn;

		if(!this.nFreqCount[0])
		{
			this.nFreqCount[0] = this.nFreqTimer[0].W + 1;
			if(++this.nDutyCount[0] >= 0x10)
			{
				this.nDutyCount[0] = 0;
				if(this.bInvert & 1)
				{
					if(this.nFreqTimer[0].W <= this.nInvertFreqCutoff)
						this.bDoInvert |= 1;
					else
						this.bDoInvert &= 2;
				}
				else
					this.bDoInvert &= 2;
			}
		}
    	if(!this.nFreqCount[1])
    	{
			this.nFreqCount[1] = this.nFreqTimer[1].W + 1;
			if(++this.nDutyCount[1] >= 0x10)
			{
				this.nDutyCount[1] = 0;
				if(this.bInvert & 2)
				{
					if(this.nFreqTimer[1].W <= this.nInvertFreqCutoff)
						this.bDoInvert |= 2;
					else
						this.bDoInvert &= 1;
				}
				else
					this.bDoInvert &= 1;
			}
		}
	}
}

inline void Wave_Squares_Mix_Mono(int * mix,int downsample)
{
	*mix += (this.nMixL / downsample);
	this.nMixL = 0;
}
	
inline void Wave_Squares_Mix_Stereo(int * mixL,int * mixR,int downsample)
{
	*mixL += (this.nMixL / downsample);
	*mixR += (this.nMixR / downsample);

	this.nMixL = this.nMixR = 0;
}

#undef this
