/* 
 * Copyright (C) 2004      Disch
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
 */

//////////////////////////////////////////////////////////////////////////
//
//  Wave_VRC6.h
//

struct VRC6PulseWave
{

	///////////////////////////////////
	// Frequency Control
	union TWIN		nFreqTimer;
	int			nFreqCount;
	int			nFreqInvHalt;

	///////////////////////////////////
	// Flags
	BYTE		bChannelEnabled;
	BYTE		bDigitized;

	///////////////////////////////////
	// Volume
	BYTE		nVolume;

	///////////////////////////////////
	// Duty Cycle
	BYTE		nDutyCycle;
	BYTE		nDutyCount;

	///////////////////////////////////
	// Output and Downsampling
	short		nOutputTable_L[0x10];
	short		nOutputTable_R[0x10];
	int			nMixL;
	int			nMixR;

	///////////////////////////////////
	// Inverting
	BYTE		bInvert;
	BYTE		bDoInvert;
	WORD		nInvertFreqCutoff;
};

extern struct VRC6PulseWave mWave_VRC6Pulse[2];
#define this mWave_VRC6Pulse[chan]

	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	inline void VRC6Pulse_DoTicks(int chan, int ticks,BYTE mix)
	{
		register int mn;

		if(!this.bChannelEnabled)
			return;

		while(ticks)
		{
			mn = min(this.nFreqCount,ticks);
			ticks -= mn;

			this.nFreqCount -= mn;

			if(mix && (this.nDutyCount <= this.nDutyCycle))
			{
				this.nMixL += this.nOutputTable_L[this.nVolume] * mn;
				this.nMixR += this.nOutputTable_R[this.nVolume] * (this.bDoInvert ? -mn : mn);
			}

			if(this.nFreqCount > 0) continue;
			this.nFreqCount = this.nFreqTimer.W + 1;

			if(!this.bDigitized)
			{
				this.nDutyCount = (this.nDutyCount + 1) & 0x0F;
				if(this.nDutyCount == 10)
				{
					this.bDoInvert = this.bInvert;
					if(this.nInvertFreqCutoff < this.nFreqTimer.W)
						this.bDoInvert = 0;
				}
			}
			else
			{
				this.bDoInvert = this.bInvert;
				if(this.nInvertFreqCutoff < this.nFreqTimer.W)
					this.bDoInvert = 0;
			}
		}
	}

	inline void Wave_VRC6Pulse_Mix_Mono(int chan, int * mix,int downsample)
	{
		mix += (this.nMixL / downsample);
		this.nMixL = 0;
	}
	
	inline void Wave_VRC6Pulse_Mix_Stereo(int chan, int * mixL,int * mixR,int downsample)
	{
		*mixL += (this.nMixL / downsample);
		*mixR += (this.nMixR / downsample);

		this.nMixL = this.nMixR = 0;
	}

#undef this

struct VRC6SawWave
{

	///////////////////////////////////
	// Frequency Control
	union TWIN		nFreqTimer;
	int			nFreqCount;

	///////////////////////////////////
	// Flags
	BYTE		bChannelEnabled;

	///////////////////////////////////
	// Phase Accumulator
	BYTE		nAccumRate;
	BYTE		nAccum;
	BYTE		nAccumStep;

	///////////////////////////////////
	// Output and Downsampling
	short		nOutputTable_L[0x20];
	short		nOutputTable_R[0x20];
	int			nMixL;
	int			nMixR;
	
	///////////////////////////////////
	// Inverting
	BYTE		bInvert;
	BYTE		bDoInvert;
	WORD		nInvertFreqCutoff;
};

extern struct VRC6SawWave mWave_VRC6Saw;
#define this mWave_VRC6Saw

	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////
	//  Functions
	///////////////////////////////////////////////////////////////////
	///////////////////////////////////////////////////////////////////

	inline void VRC6Saw_DoTicks(int ticks,BYTE mix)
	{
		register int mn;

		if(!this.bChannelEnabled)
			return;

		while(ticks)
		{
			mn = min(this.nFreqCount,ticks);
			ticks -= mn;

			this.nFreqCount -= mn;

			if(mix)
			{
				this.nMixL += this.nOutputTable_L[this.nAccum >> 3] * mn;
				this.nMixR += this.nOutputTable_R[this.nAccum >> 3] * (this.bDoInvert ? -mn : mn);
			}

			if(this.nFreqCount > 0) continue;
			this.nFreqCount = this.nFreqTimer.W + 1;

			this.nAccumStep++;
			if(this.nAccumStep == 14)
			{
				this.nAccumStep = 0;
				this.nAccum = 0;
				this.bDoInvert = this.bInvert;
				if(this.nInvertFreqCutoff < this.nFreqTimer.W)
					this.bDoInvert = 0;
			}
			else if(!(this.nAccumStep & 1))
				this.nAccum += this.nAccumRate;
		}
	}

	inline void Wave_VRC6Saw_Mix_Mono(int * mix,int downsample)
	{
		*mix += (this.nMixL / downsample);
		this.nMixL = 0;
	}
	
	inline void Wave_VRC6Saw_Mix_Stereo(int * mixL,int * mixR,int downsample)
	{
		*mixL += (this.nMixL / downsample);
		*mixR += (this.nMixR / downsample);
        
		this.nMixL = this.nMixR = 0;
	}
	
#undef this
